﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------

-- basic layout

nUI_MicroMenuLayout =
{
	btn_size = 55,
	gap      = -2,
	anchor   = "TOP",
	xOfs     = 0,
	yOfs     = -2,
};

-- the MicroMenu bar frame

nUI_MicroMenu = CreateFrame( "Frame", "nUI_MicroMenu", nUI_Console );

nUI_MicroMenu:SetBackdrop(
	{
		bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
		edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
		tile     = true, 
		tileSize = 1, 
		edgeSize = 12, 
		insets   = {left = 0, right = 0, top = 0, bottom = 0},
	}
);

nUI_MicroMenu:SetBackdropColor( 0, 0, 0, 0.35 );

-- button initialization

nUI_MicroMenu.Buttons = {};

nUI_MicroMenu.Buttons[1] = _G["CharacterMicroButton"];
nUI_MicroMenu.Buttons[2] = _G["SpellbookMicroButton"];
nUI_MicroMenu.Buttons[3] = _G["TalentMicroButton"];
nUI_MicroMenu.Buttons[4] = _G["QuestLogMicroButton"];
nUI_MicroMenu.Buttons[5] = _G["SocialsMicroButton"];
nUI_MicroMenu.Buttons[6] = _G["LFGMicroButton"];
nUI_MicroMenu.Buttons[7] = _G["MainMenuMicroButton"];
nUI_MicroMenu.Buttons[8] = _G["HelpMicroButton"];

for i in pairs( nUI_MicroMenu.Buttons ) do
	
	button = nUI_MicroMenu.Buttons[i];
	
	button:SetParent( nUI_MicroMenu );

	button.nUI_CachedClearAllPoints = button.ClearAllPoints;
	button.nUI_CachedSetAllPoints   = button.SetAllPoints;
	button.nUI_CachedSetPoint       = button.SetPoint;

	button.ClearAllPoints = function() end;
	button.SetAllPoints   = function() end;
	button.SetPoint       = function() end;
end

-- manage the layout of the MicroMenu Bar

nUI_MicroMenu:SetScript( "OnEvent", function() nUI_MicroMenu:layoutFrame(); end );
nUI_MicroMenu:RegisterEvent( "VARIABLES_LOADED" );
nUI_MicroMenu:RegisterEvent( "DISPLAY_SIZE_CHANGED" );

nUI_MicroMenu.layoutFrame = function()
				
	nUI:setScale();
	
	if nUI_MicroMenu.fCache then nUI:lockFrame( nUI_MicroMenu, false ); end
	
	local scale       = nUI_Console:GetEffectiveScale();
	local btn_size    = nUI_MicroMenuLayout.btn_size * scale;
	local btn_gap     = 0;
	local width       = 0;
	local last_button;
	
	nUI_MicroMenu:SetScale( 1.0 / scale );	
	nUI_MicroMenu:SetFrameStrata( nUI_Console:GetFrameStrata() );
	nUI_MicroMenu:SetFrameLevel( nUI_Console:GetFrameLevel()+2 );
	
	for i in pairs( nUI_MicroMenu.Buttons ) do
		
		local button = nUI_MicroMenu.Buttons[i];
		
		button:SetFrameStrata( nUI_MicroMenu:GetFrameStrata() );
		button:SetFrameLevel( nUI_MicroMenu:GetFrameLevel()+1 );
		button:SetScale( btn_size / 36 );
		
		width = width + button:GetWidth() * button:GetScale() + (i > 1 and btn_gap or 0);

		button:nUI_CachedClearAllPoints();
		
		if i == 1 then button:nUI_CachedSetPoint( "BOTTOMLEFT", nUI_MicroMenu, "BOTTOMLEFT", btn_size * 0.1, btn_size * 0.1 );
		else button:nUI_CachedSetPoint( "TOPLEFT", last_button, "TOPRIGHT", btn_gap, 0 );
		end

		last_button = button;
	end		
	
	nUI_MicroMenu:SetHeight( btn_size * 1.1 );
	nUI_MicroMenu:SetWidth( width + btn_size * 0.1 );
	nUI_MicroMenu:SetPoint( nUI_MicroMenuLayout.anchor, nUI_Console, "TOP", nUI_MicroMenuLayout.xOfs * nUI.scale, nUI_MicroMenuLayout.yOfs * nUI.scale );
	
	nUI:lockFrame( nUI_MicroMenu, true, "Micro-Menu" );
	
end
