﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------

-- the possess bar frame

nUI_PossessBar = CreateFrame( "Frame", "nUI_PossessBar", nUI_Dashboard, "SecureStateHeaderTemplate" );

nUI_PossessBar:SetBackdrop(
	{
		bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
		edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
		tile     = true, 
		tileSize = 1, 
		edgeSize = 12, 
		insets   = {left = 0, right = 0, top = 0, bottom = 0},
	}
);

nUI_PossessBar:SetBackdropColor( 0, 0, 0, 0.35 );
nUI_PossessBar:SetAttribute( "showstates", 3 );
nUI_PetBarAnchor:SetAttribute( "addchild", nUI_PossessBar );

-- button initialization

nUI_PossessBar.Buttons = {};

for i=1,10 do

	local button    = CreateFrame( "CheckButton", "$parent_Button"..i, nUI_PossessBar, "PossessButtonTemplate" );	

	nUI_PossessBar.Buttons[i] = button;
	
	button:SetID( i );
	button:SetParent( nUI_PossessBar );
	button:SetNormalTexture( "" );
	
	button.nUI_CachedClearAllPoints = button.ClearAllPoints;
	button.nUI_CachedSetAllPoints   = button.SetAllPoints;
	button.nUI_CachedSetPoint       = button.SetPoint;

	button.ClearAllPoints = function() end;
	button.SetAllPoints   = function() end;
	button.SetPoint       = function() end;

	nUI_PossessBar:SetAttribute( "addchild", button );
end

-- manage the layout of the Possess Bar

nUI_PossessBar:SetScript( "OnEvent", 
	function() 

		if event == "PLAYER_AURAS_CHANGED" 
		or event == "UPDATE_BONUS_ACTIONBAR"
		then		
			for i=1, NUM_POSSESS_SLOTS do				
				nUI_PossessBar.updateState( i );				
			end
		else		
			nUI_PossessBar:layoutFrame(); 
		end
	end 
);

nUI_PossessBar:RegisterEvent( "VARIABLES_LOADED" );
nUI_PossessBar:RegisterEvent( "DISPLAY_SIZE_CHANGED" );
nUI_PossessBar:RegisterEvent( "PLAYER_AURAS_CHANGED" );
nUI_PossessBar:RegisterEvent( "UPDATE_BONUS_ACTIONBAR" );

nUI_PossessBar.updateState = function( index )
	
	button = nUI_PossessBar.Buttons[index];
	
	texture, name = GetPossessInfo( index );

	button.layers.icon:SetTexture( texture );	
	button.layers.icon:SetVertexColor( 1.0, 1.0, 1.0 );
	
	button:SetChecked( 0 );
	
end

nUI_PossessBar.layoutFrame = function()
				
	nUI:setScale();
				
	local btn_size    = nUI_PetBarLayout.btn_size * nUI.scale;
	local btn_gap     = nUI_PetBarLayout.gap * nUI.scale;
	local width       = 0;
	local last_button;
	
	nUI_PossessBar:SetScale( 1.0 / nUI.scale );	
	nUI_PossessBar:SetFrameStrata( nUI_Dashboard:GetFrameStrata() );
	nUI_PossessBar:SetFrameLevel( nUI_Dashboard:GetFrameLevel()+3 );
	
	for i in pairs( nUI_PossessBar.Buttons ) do
		
		local button = nUI_PossessBar.Buttons[i];

		button:SetFrameStrata( nUI_PossessBar:GetFrameStrata() );
		button:SetFrameLevel( nUI_PossessBar:GetFrameLevel()+1 );
		button:SetScale( btn_size / button:GetHeight() );

		nUI:initActionButton( button );

		if i > NUM_POSSESS_SLOTS then
			
			button:Hide();
			
		else
			
			width = width + button:GetWidth() * button:GetEffectiveScale() + (i > 1 and btn_gap or 0);
	
			button:nUI_CachedClearAllPoints();
			
			if i == 1 then button:nUI_CachedSetPoint( "TOPLEFT", nUI_PossessBar, "TOPLEFT", btn_size * 0.05, -btn_size * 0.05 );
			else button:nUI_CachedSetPoint( "TOPLEFT", last_button, "TOPRIGHT", btn_gap, 0 );
			end
		
			nUI_PossessBar.updateState( i );
	
			button.layers.cooldown:Hide();
						
		end
		
		last_button = button;
	end		

	nUI_PossessBar:SetHeight( btn_size * 1.1 );
	nUI_PossessBar:SetWidth( width + btn_size * 0.1 );
	
	nUI_PossessBar:ClearAllPoints();
	nUI_PossessBar:SetPoint( "TOP", nUI_PetBar, "TOP", 0, 0 );
	
end