﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------
	
-------------------------------------------------------------------------------
-- local cache

local CreateFrame             = CreateFrame;
local ToggleDropDownMenu      = ToggleDropDownMenu;
local UnitIsUnit              = UnitIsUnit;
local UnitIsPlayer            = UnitIsPlayer;
local UnitInParty             = UnitInParty;
local UnitLevel               = UnitLevel;
local UnitReaction            = UnitReaction;
local UnitPowerType           = UnitPowerType;
local UnitExists              = UnitExists;
local UnitName                = UnitName;
local UnitClass               = UnitClass;
local UnitIsPVP               = UnitIsPVP;
local UnitClassification      = UnitClassification;
local UnitIsPVPFreeForAll     = UnitIsPVPFreeForAll;
local UnitIsVisible           = UnitIsVisible;
local UnitIsFriend            = UnitIsFriend;
local UnitPlayerControlled    = UnitPlayerControlled;
local UnitPlayerOrPetInParty  = UnitPlayerOrPetInParty;
local UnitPlayerOrPetInRaid   = UnitPlayerOrPetInRaid;
local UnitIsTapped            = UnitIsTapped;
local UnitIsTappedByPlayer    = UnitIsTappedByPlayer;
local UnitGUID                = UnitGUID;
local UnitIsPVPSanctuary      = UnitIsPVPSanctuary;
local SetPortraitTexture      = SetPortraitTexture;
local UnitIsConnected         = UnitIsConnected;
local UnitIsAFK               = UnitIsAFK;
local UnitIsDead              = UnitIsDead;
local UnitIsDeadOrGhost       = UnitIsDeadOrGhost;
local GetPetHappiness         = GetPetHappiness;
local HasPetUI                = HasPetUI;
local IsResting               = IsResting;
local GetRaidTargetIndex      = GetRaidTargetIndex;
local UnitIsPartyLeader       = UnitIsPartyLeader;
local NotifyInspect           = NotifyInspect;
local GetTime                 = GetTime;

nUI.Class.Unit  = nUI.Class:CreatePrototype();
local prototype = nUI.Class.Unit.prototype;

-------------------------------------------------------------------------------
-- initialize click casting registration

if not ClickCastFrames then
	ClickCastFrames = {};
end

nUI.ClickCastFrames = {};

-------------------------------------------------------------------------------
-- spec browsing queue handling
--
-- requesting talent page info on remote inspections carries a delay. You have
-- to notify inspect on the unit, then wait for the INSPECT_TALENT_READY event
-- to read the talent info. This code creates a single threaded FIFO queue of
-- inspection requests to avoid collisions between myriad unit frames updating
-- in a short period of time (i.e. and AV raid)

-------------------------------------------------------------------------------

function prototype:init( parent, id, layout )
	
	self.parent         = parent;
	self.id             = id;
	self.rid            = 1;
	
	self:layoutFrame( layout, true );
	
end

-------------------------------------------------------------------------------

function prototype:layoutFrame( layout, is_init )

	if not is_init then 
		
		UnregisterUnitWatch( self.frame );
		self:unregisterEvents(); 
		self.frame:Hide();
		if self.border then self.border:Hide(); end

		for i in pairs( self.ClickCastFrames ) do
			
			local frame = self.ClickCastFrames[i]
			
			ClickCastFrames[frame]     = false;
			nUI.ClickCastFrames[frame] = false;
			
		end		
	end
	
	local nUI_scale      = nUI.scale;
	local unit_id        = layout.unit_id;
	local dX             = (layout.Size.width - 6) * nUI_scale;
	local dY             = (layout.Size.height - 6) * nUI_scale;
	local info_box_size  = math.floor( dY / 3 );
	local dL             = dY - info_box_size * 2;
	local portrait;		
	
	self.layout          = layout;
	self.has_model       = false;
	self.is_shown        = true;
	self.talent_queue    = 0;
	self.Cache           = {};
	self.unit_name       = UnitName( unit_id ) or "";	
	self.frame_name      = (layout.name or "nUI_Unit").."_"..(unit_id or self.id);
	self.Timers          = {};
	self.ClickCastFrames = {};
	self.Timers.refresh  = 0;
	self.Timers.cam_fix  = 0;	
	
	-- calculate dimentions for the boxes based on the unit options
	
	if layout.Size.compact == "yes" then
		
		dL = dY;
		info_box_size = dY;
		self.compact  = true;
		
	elseif layout.show_detail == "no" and layout.StatusBars.visible == "no" then
		
		dL            = dY;
		info_box_size = 0;
		
	elseif layout.show_detail ~= "no" and (layout.Size.two_line == "yes" or layout.show_class == "no") then
		
		info_box_size = math.floor( dY * 0.5 ) + 2;
		dL            = dY - info_box_size;
		
	end
	
	self.dY = dY;
	self.dX = dX;
	
	-- create the base frame and event handler

	self.frame = self:createUnitFrame( self.frame_name, self.parent, false );
	self.frame.class    = self;
	self.frame.is_shown = true;
	self.frame.unit     = unit_id;

	self.frame:EnableMouse( false );
	self.frame:SetFrameStrata( "LOW" );
	self.frame:SetFrameLevel( 0 );
	self.frame:SetScale( 1.0 / nUI_scale );
	self.frame:SetWidth( layout.Size.width * nUI_scale );
	self.frame:SetHeight( layout.Size.height * nUI_scale );
	self.frame:SetPoint( layout.Origin.anchor, self.parent, "BOTTOM", layout.Origin.xOfs * nUI_scale, layout.Origin.yOfs * nUI_scale );
	
	RegisterUnitWatch( self.frame );
	
	-- create a layer for the unit popup

	self.frame.popup = CreateFrame( "Frame", "$parent_Popup", self.frame, "UIDropDownMenuTemplate" );
	self.frame.popup.class = self;

	self.frame.popup:SetFrameLevel( 9 );
	self.frame.popup:SetPoint( "CENTER", self.frame, "CENTER", 0, 0 );
	
	self.frame.popup.texture = self.frame.popup:CreateTexture( "$parentTexture" );
	self.frame.popup.texture:SetAllPoints( self.frame.popup );
	self.frame.popup.texture:SetTexture( 0, 0, 0, 0.75 );

	self.frame.popup:Hide();

	UIDropDownMenu_Initialize( 
		self.frame.popup, 
		function () self:updateUnitPopup();	end, 
		"MENU" 
	);

	HideDropDownMenu( 1 );
	
	-- unit ID used in raids and groups
	
	if layout.id then self.frame:SetID( layout.id ); end
	
	-- create an optional border to outline where the unit is located
	
	if layout.show_border == "yes" then
		
		self.border = CreateFrame( "Frame",  self.frame_name.."_Border", self.parent );
		self.border:SetFrameStrata( "LOW" );
		self.border:SetScale( 1.0 / nUI_scale );
		self.border:SetAllPoints( self.frame );
		self.border:SetFrameLevel( 0 );

		self.border:SetBackdrop(
			{
				bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
				edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
				tile     = true, 
				tileSize = 1, 
				edgeSize = 12, 
				insets   = {left = 0, right = 0, top = 0, bottom = 0},
			}
		);

		self.border:SetBackdropColor( 0, 0, 0, 0.35 );
		self.border:Show();
	end
	
	-- if we're showing the portrait, then create the frame for that
	
	if not self.compact and layout.Portrait.visible ~= "no" then
		
		dX = dX - dY;
		
		self.has_model   = true;
		self.frame.model = self:createUnitFrame( "$parent_Model", self.frame, true );

		self.frame.model:SetFrameLevel( 3 );
		self.frame.model:SetBackdrop(
			{
				bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
				edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
				tile     = true, 
				tileSize = 1, 
				edgeSize = 10, 
				insets   = {left = 2, right = 2, top = 2, bottom = 2},
			}
		);
		self.frame.model:SetBackdropColor( 0, 0, 0, 0.75 );
		self.frame.model:SetWidth( dY );
		self.frame.model:SetHeight( dY );
		
		if layout.orientation == "left" then
			self.frame.model:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", 2, -2 );
		else
			self.frame.model:SetPoint( "TOPRIGHT", self.frame, "TOPRIGHT", -2, -2 );
		end
		
		-- the actual model animation or still portrait as the case may be
		
		self.Cache.portrait = CreateFrame( "Frame", "$parentPortrait", self.frame.model );
		
		self.Cache.portrait:SetFrameLevel( 4 );
		self.Cache.portrait:SetWidth( self.dY - 6 );
		self.Cache.portrait:SetHeight( self.dY - 6 );
		self.Cache.portrait:SetPoint( "TOPLEFT", self.frame.model, "TOPLEFT", 3, -3 );

		self.Cache.portrait.texture = self.Cache.portrait:CreateTexture( "$parentTexture" );
		self.Cache.portrait.texture:SetAllPoints( self.Cache.portrait );
		self.Cache.portrait:SetAlpha( 0 );

		self.Cache.anim = CreateFrame( "PlayerModel", "$parentAnim", self.frame.model );
		
		self.Cache.anim:SetFrameLevel( 4 );
		self.Cache.anim:SetWidth( self.dY - 6 );
		self.Cache.anim:SetHeight( self.dY - 6 );
		self.Cache.anim:SetPoint( "TOPLEFT", self.frame.model, "TOPLEFT", 3, -3 );
		self.Cache.anim:SetScript( "OnShow", function() self.Cache.anim:SetCamera( 0 ); end );
		self.Cache.anim:SetUnit( unit_id );
		self.Cache.anim:SetAlpha( 0 );
		
		-- a text box for describing the unit
		
		self.frame.model.info = CreateFrame( "Frame", "$parentType", self.frame.model );

		self.frame.model.info:SetFrameLevel( 5 );

		self.frame.model.info:SetBackdrop(
			{
				bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
				edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
				tile     = true, 
				tileSize = 1, 
				edgeSize = 10, 
				insets   = {left = 2, right = 2, top = 2, bottom = 2},
			}
		);
		self.frame.model.info:SetBackdropBorderColor( 1, 1, 1, 0.25 );
		self.frame.model.info:SetBackdropColor( 0, 0, 0, 0.5 );
		self.frame.model.info:SetWidth( dY+1 );
		self.frame.model.info:SetHeight( dY / 4 );
		self.frame.model.info:SetPoint( "BOTTOM", self.frame.model, "BOTTOM", 0, 0 );
		self.frame.model.info:SetAlpha( 0 );

		self.frame.model.info.text = self.frame.model.info:CreateFontString( "$parentText", "OVERLAY" );
		self.frame.model.info.text:SetPoint( "CENTER", self.frame.model.info, "CENTER", 0, 0 );
		self.frame.model.info.text:SetFont( "Fonts\\ARIALN.TTF", max( ((dY / 4 - 6) * 0.9), 4), "OUTLINE" );
		self.frame.model.info.text:SetTextColor( 1, 1, 1, 1 );
		self.frame.model.info.text:SetJustifyH( "CENTER" );
		self.frame.model.info.text:SetJustifyV( "MIDDLE" );

		-- overlay frame for elites, rares and world bosses

		if layout.elite_icon == "yes" then
			self.frame.model.elite = CreateFrame( "Frame", "$parentIcon", self.frame.model );
			self.frame.model.elite:SetWidth( 256 );
			self.frame.model.elite:SetHeight( 256 );
			self.frame.model.elite:SetScale( (dY - 6) * 0.009 );

			self.frame.model.elite.texture = self.frame.model.elite:CreateTexture( "$parentTexture" );
			self.frame.model.elite.texture:SetAllPoints( self.frame.model.elite );

			if layout.orientation == "left" then
				self.frame.model.elite:SetPoint( "CENTER", self.frame.model, "CENTER", -18, -10 );
			else
    			self.frame.model.elite:SetPoint( "CENTER", self.frame.model, "CENTER", 18, -10 );
			end

			-- depending on whether the model is square or round, we either overlay the info
			-- over the elite icon (round models) or visa versa (square models)
			
			if layout.Portrait.style == "square" then
				self.frame.model.elite:SetFrameLevel( 6 );
				self.frame.model.info:SetFrameLevel( 5 );
			else
				self.frame.model.elite:SetFrameLevel( 5 );
				self.frame.model.info:SetFrameLevel( 6 );
			end            
		end

		-- combat feedback system

		self.frame.model.feedback = CreateFrame( "Frame", "$parent_Feedback", self.frame.model );
		self.frame.model.feedback.text = self.frame.model.feedback:CreateFontString( "$parentText", "OVERLAY" );
		self.frame.model.feedback.text:SetWidth( self.frame.model:GetWidth() );
		self.frame.model.feedback.text:SetHeight( self.frame.model:GetHeight() );
		self.frame.model.feedback.text:SetPoint( "CENTER", self.frame.model, "CENTER", -1, 0 );
		self.frame.model.feedback.text:SetFont( "Fonts\\skurri.ttf", self.dY * 0.29, "THICKOUTLINE" );
		self.frame.model.feedback.text:SetJustifyH( "CENTER" );
		self.frame.model.feedback.text:SetJustifyV( "MIDDLE" );
		self.frame.model.feedback:SetFrameLevel( 8 );

		-- combat and resting status
		
		self.frame.model.status = CreateFrame( "Frame", "$parent_Status", self.frame.model );
		self.frame.model.status:SetWidth( dL+2 );
		self.frame.model.status:SetHeight( dL+2 );
		self.frame.model.status:SetFrameLevel( 5 );
		self.frame.model.status:SetAlpha( 0 );
		
		if layout.orientation == "left" then
			self.frame.model.status:SetPoint( "TOPLEFT", self.frame.model, "TOPLEFT", 0, 0 );
		else
			self.frame.model.status:SetPoint( "TOPRIGHT", self.frame.model, "TOPRIGHT", 0, 0 );
		end
		
		self.frame.model.status.texture = self.frame.model.status:CreateTexture( "$parentTexture" );
		self.frame.model.status.texture:SetAllPoints( self.frame.model.status );
		self.frame.model.status.texture:SetTexture( "Interface\\AddOns\\nUI\\Units\\Art\\nUI_StateIcon.blp" );
		
	end

	-- frame for the name of the mob in the unit
	
	self.frame.label = self:createUnitFrame( "$parent_Label", self.frame, true );

	self.frame.label:SetFrameLevel( self.compact and 5 or 3 );
	self.frame.label:SetBackdrop(
		{
			bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
			edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
			tile     = true, 
			tileSize = 1, 
			edgeSize = 10, 
			insets   = {left = 2, right = 2, top = 2, bottom = 2},
		}
	);
	self.frame.label:SetBackdropColor( 0, 0, 0, self.compact and 0 or 0.75 );
    self.frame.label:SetWidth( dX+1 );
	self.frame.label:SetHeight( dL+2 );
	
	self.frame.label.text = self.frame.label:CreateFontString( "$parentText", "OVERLAY" );
	self.frame.label.text:SetFont( "Fonts\\FRIZQT__.TTF", (dL+2) * (self.compact and 0.35 or 0.475), "OUTLINE" );
	self.frame.label.text:SetTextColor( 1, 0.8196079, 0, 1 );
	self.frame.label.text:SetJustifyH( "CENTER" );
	self.frame.label.text:SetJustifyV( "MIDDLE" );

	if self.compact then
		self.frame.label.text:SetPoint( "TOP", self.frame.label, "TOP", 0, -dL/4 );
	else
		self.frame.label.text:SetPoint( "CENTER", self.frame.label, "CENTER", 0, 0 );
	end
	
	if layout.orientation == "left" then
		self.frame.label:SetPoint( "TOPRIGHT", self.frame, "TOPRIGHT", -2, -2 );
	else
		self.frame.label:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", 2, -2 );
	end		

	-- a text layer for displaying the unit's raid group
	
	if layout.show_group == "yes" then

		self.frame.raid  = CreateFrame( "Frame", "$parent_RaidGroup", self.frame.label );
		self.frame.group = self.frame.raid:CreateFontString();
		
		self.frame.raid:SetBackdrop(
			{
				bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
				edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
				tile     = true, 
				tileSize = 1, 
				edgeSize = 10, 
				insets   = {left = 2, right = 2, top = 2, bottom = 2},
			}
		);
		self.frame.raid:SetBackdropBorderColor( 1, 1, 1, 0.5 );
		self.frame.raid:SetBackdropColor( 0, 0, 0, 0.5 );
		self.frame.raid:SetFrameLevel( 5 );
		self.frame.raid:SetHeight( dL+2 );
		self.frame.raid:SetWidth( info_box_size );
		self.frame.raid:SetAlpha( 0 );
		
		self.frame.group:SetFont( "Fonts\\ARIALN.TTF", info_box_size * 0.6, "OUTLINE" );
		self.frame.group:SetTextColor( 1, 1, 1, 1 );
		self.frame.group:SetJustifyV( "MIDDLE" );
		self.frame.group:SetJustifyH( "CENTER" );
		self.frame.group:SetPoint( "CENTER", self.frame.raid, "CENTER", 0, 0 );
		
		if layout.orientation ~= "left" then
			if self.frame.model then self.frame.raid:SetPoint( "TOPLEFT", self.frame.model, "TOPLEFT", 0, 0 );
			else self.frame.raid:SetPoint( "TOPLEFT", self.frame.label, "TOPRIGHT", 2, 0 );
			end
		else
			if self.frame.model then self.frame.raid:SetPoint( "TOPRIGHT", self.frame.model, "TOPRIGHT", 0, 0 );
			else self.frame.raid:SetPoint( "TOPRIGHT", self.frame.label, "TOPLEFT", -2, 0 );
			end
		end
	end
	
	-- unit PvP/FFA frame
	
	self.frame.pvp = CreateFrame( "Frame", "$parent_PVP", self.frame );
	self.frame.pvp:SetHeight( self.compact and dL*2/3 or dL );
	self.frame.pvp:SetWidth( self.compact and dL*2/3 or dL );
	self.frame.pvp:SetFrameLevel( self.compact and 6 or 4 );
	self.frame.pvp:SetAlpha( 0 );
	
	self.frame.pvp.texture = self.frame.pvp:CreateTexture( "$parentTexture" );
	self.frame.pvp.texture:SetAllPoints( self.frame.pvp );
	
	if layout.orientation == "left" then
		
		self.frame.pvp:SetPoint( "CENTER", self.frame.label, "RIGHT", 0, 0 );
		
	else
		
		self.frame.pvp:SetPoint( "CENTER", self.frame.label, "LEFT", 0, 0 );
		
	end
	
	-- frame for indicating group leader/assist status
	
	self.frame.leader = CreateFrame( "Frame", "$parent_Leader", self.frame );
	self.frame.leader:SetFrameLevel( 6 );
	self.frame.leader:SetHeight( self.compact and (dL/3) or (dL/2) );
	self.frame.leader:SetWidth( self.compact and (dL/3) or (dL/2) );
	self.frame.leader:SetPoint( "TOP", self.frame.label, "TOP", 0, 0 );
	self.frame.leader:SetAlpha( 0 );
	
	self.frame.leader.texture = self.frame.leader:CreateTexture( "$parentTexture" );
	self.frame.leader.texture:SetAllPoints( self.frame.leader );
	
	-- frame for indicating master looter status
	
	self.frame.master_looter = CreateFrame( "Frame", "$parent_MasterLooter", self.frame );
	self.frame.master_looter:SetFrameLevel( 6 );
	self.frame.master_looter:SetHeight( self.compact and (dL/3) or (dL/2) );
	self.frame.master_looter:SetWidth( self.compact and (dL/3) or (dL/2) );
	self.frame.master_looter:SetAlpha( 0 );

	if layout.orientation == "left" then
		self.frame.master_looter:SetPoint( "RIGHT", self.frame.leader, "LEFT", 0, 0 );
	else
		self.frame.master_looter:SetPoint( "LEFT", self.frame.leader, "RIGHT", 0, 0 );
	end
	
	self.frame.master_looter.texture = self.frame.master_looter:CreateTexture( "$parentTexture" );
	self.frame.master_looter.texture:SetAllPoints( self.frame.master_looter );
	self.frame.master_looter.texture:SetTexture( "Interface\\GroupFrame\\UI-Group-MasterLooter.blp" );
	self.frame.master_looter.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
	
	-- frame for indicating main tank/off-tank status
	
	self.frame.tank = CreateFrame( "Frame", "$parent_TankIcon", self.frame );
	self.frame.tank:SetFrameLevel( 6 );
	self.frame.tank:SetHeight( self.compact and (dL/3) or (dL/2) );
	self.frame.tank:SetWidth( self.compact and (dL/3) or (dL/2) );
	self.frame.tank:SetAlpha( 0 );

	if layout.orientation == "left" then
		self.frame.tank:SetPoint( "LEFT", self.frame.leader, "RIGHT", 0, 0 );
	else
		self.frame.tank:SetPoint( "RIGHT", self.frame.leader, "LEFT", 0, 0 );
	end
	
	self.frame.tank.texture = self.frame.master_looter:CreateTexture( "$parentTexture" );
	self.frame.tank.texture:SetAllPoints( self.frame.master_looter );

	-- raid icon
	
	self.frame.raid_icon = CreateFrame( "Frame", "$parent_RaidIcon", self.frame );
	self.frame.raid_icon:SetFrameLevel( 7 );
	self.frame.raid_icon:SetHeight( dL );
	self.frame.raid_icon:SetWidth( dL );
	self.frame.raid_icon:SetAlpha( 0 );

	self.frame.raid_icon.texture = self.frame.raid_icon:CreateTexture( "$parentTexture" );
	self.frame.raid_icon.texture:SetHeight( dL );
	self.frame.raid_icon.texture:SetWidth( dL );
	self.frame.raid_icon.texture:SetAllPoints( self.frame.raid_icon );
	self.frame.raid_icon.texture:SetTexture( "Interface\\TargetingFrame\\UI-RaidTargetingIcons" );
	
	if layout.orientation == "left" then			
		self.frame.raid_icon:SetPoint( "CENTER", self.frame.label, "LEFT", 0, 0 );			
	else
		self.frame.raid_icon:SetPoint( "CENTER", self.frame.label, "RIGHT", 0, 0 );			
	end
	
	-- unit casting/channeling bar
	
	if not self.compact and layout.CastingBars.visible ~= "no" then

		local Cy = self.frame.label:GetHeight() - 6;
		local Hx = self.frame.label:GetWidth() - 6;
		
		if layout.CastingBars.latency == "yes" then Cy = Cy - 3; end

		layout.CastBar.height = Cy;
		layout.CastBar.width  = Hx;				
		self.frame.casting    = nUI.Class.UnitBar_Casting:new( self.frame.label, self.id+self.rid, layout.CastBar, unit_id );
		self.rid              = self.rid + (self.frame.casting.rid or 1);					
		
		self.frame.casting:SetPoint( "TOPLEFT", self.frame.label, "TOPLEFT", 3, -3 );
		
		layout.ChannelBar.height = Cy;
		layout.ChannelBar.width  = Hx;				
		self.frame.channel       = nUI.Class.UnitBar_Channeling:new( self.frame.label, self.id+self.rid, layout.ChannelBar, unit_id );
		self.rid                 = self.rid + (self.frame.channel.rid or 1);					
		
		self.frame.channel:SetPoint( "TOPLEFT", self.frame.label, "TOPLEFT", 3, -3 );
	
		-- unit latency bar... cannot exist without a casting bar
		
		if layout.CastingBars.latency == "yes" then
			
			layout.LatencyBar.height = 3;
			layout.LatencyBar.width  = Hx;	
			self.frame.latency       = nUI.Class.UnitBar_Latency:new( self.frame.label, self.id+self.rid, layout.LatencyBar, unit_id );
			self.rid                 = self.rid + (self.frame.latency.rid or 1);					
			
			self.frame.latency:SetPoint( "TOPLEFT", layout.CastBar.name, "BOTTOMLEFT", 0, 0 );

		end
	end
	
	-- unit detail information
	
	if not self.compact and layout.show_detail ~= "no" then

		dX = dX - info_box_size + 2;
		
		-- frame for displaying the mob's level (or skull for bosses) and class
	
		self.frame.level = self:createUnitFrame( "$parent_Level", self.frame, true );

		self.frame.level:SetFrameLevel( 3 );
		self.frame.level:SetBackdrop(
			{
				bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
				edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
				tile     = true, 
				tileSize = 1, 
				edgeSize = 10, 
				insets   = {left = 2, right = 2, top = 2, bottom = 2},
			}
		);
		self.frame.level:SetBackdropColor( 0, 0, 0, 0.75 );
		self.frame.level:SetWidth( info_box_size );
		self.frame.level:SetHeight( info_box_size  );
		
		self.frame.level.text = self.frame.level:CreateFontString( "$parent_LevelText", "OVERLAY" );
		self.frame.level.text:SetPoint( "CENTER", self.frame.level, "CENTER", 0.75, 0.75 );
		self.frame.level.text:SetFont( "Fonts\\ARIALN.TTF", info_box_size * 0.6, "OUTLINE" );
		self.frame.level.text:SetTextColor( 1, 1, 1, 1 );
		self.frame.level.text:SetJustifyH( "CENTER" );
		self.frame.level.text:SetJustifyV( "MIDDLE" );

		if layout.orientation == "left" then
			self.frame.level:SetPoint( "TOPLEFT", self.frame.label, "BOTTOMLEFT", 0, self.compact and 3 or 1 );
		else
			self.frame.level:SetPoint( "TOPRIGHT", self.frame.label, "BOTTOMRIGHT", 0, self.compact and 3 or 1 );
		end
		
		-- a frame for displaying the boss icon for those mobs that require it
		
		self.frame.level.boss = CreateFrame( "Frame", "$parent_BossIcon", self.frame.level );

		self.frame.level.boss:SetFrameLevel( 4 );
		self.frame.level.boss:SetBackdrop(
			{
				bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
				edgeFile = nil, 
				tile     = true, 
				tileSize = 1, 
				edgeSize = 10, 
				insets   = {left = 2, right = 2, top = 2, bottom = 2},
			}
		);
		self.frame.level.boss:SetBackdropColor( 0, 0, 0, 1 );
		self.frame.level.boss:SetPoint( "TOPLEFT", self.frame.level, "TOPLEFT", 3, -3 );
		self.frame.level.boss:SetPoint( "TOPRIGHT", self.frame.level, "TOPRIGHT", -3, -3 );
		self.frame.level.boss:SetPoint( "BOTTOMLEFT", self.frame.level, "BOTTOMRIGHT", 3, 3 );
		self.frame.level.boss:SetPoint( "BOTTOMRIGHT", self.frame.level, "BOTTOMRIGHT", -3, 3 );

		self.frame.level.boss.texture = self.frame.level.boss:CreateTexture( "$parentTexture" );
		self.frame.level.boss.texture:SetTexture( "Interface\\TargetingFrame\\UI-TargetingFrame-Skull" );
		self.frame.level.boss.texture:SetAllPoints( self.frame.level.boss );

		-- frame for displaying the unit's class icon (warrior, shaman, etc.)
		
		if layout.show_class ~= "no" then
			
			self.frame.unit_class = self:createUnitFrame( "$parent_Class", self.frame, true );

			self.frame.unit_class:SetFrameLevel( 3 );
			self.frame.unit_class:SetBackdrop(
				{
					bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
					edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
					tile     = true, 
					tileSize = 1, 
					edgeSize = 10, 
					insets   = {left = 2, right = 2, top = 2, bottom = 2},
				}
			);
			self.frame.unit_class:SetBackdropColor( 0, 0, 0, 0.75 );
			self.frame.unit_class:SetWidth( info_box_size );
			self.frame.unit_class:SetHeight( info_box_size );
			
			if layout.Size.two_line == "yes" then
				if layout.orientation ~= "right" then
					self.frame.unit_class:SetPoint( "LEFT", self.frame.level, "RIGHT", -1, 0 );
				else
					self.frame.unit_class:SetPoint( "RIGHT", self.frame.level, "LEFT", 1, 0 );
				end
				
				dX = dX - info_box_size + 1;
			else
				self.frame.unit_class:SetPoint( "TOPLEFT", self.frame.level, "BOTTOMLEFT", 0, 1 );
			end
						
			self.frame.unit_class.texture = self.frame.unit_class:CreateTexture( "$parentTexture" );
			self.frame.unit_class.texture:SetTexture( "Interface\\AddOns\\nUI\\Units\\Art\\nUI_Icon_Classes.blp" );
			self.frame.unit_class.texture:SetPoint( "TOPLEFT", self.frame.unit_class, "TOPLEFT", 3, -3 );
			self.frame.unit_class.texture:SetPoint( "TOPRIGHT", self.frame.unit_class, "TOPRIGHT", -3, -3 );
			self.frame.unit_class.texture:SetPoint( "BOTTOMLEFT", self.frame.unit_class, "BOTTOMLEFT", 3, 3 );
			self.frame.unit_class.texture:SetPoint( "BOTTOMRIGHT", self.frame.unit_class, "BOTTOMRIGHT", -3, 3 );
		end
	end
	
	-- create a pet happiness indicator
	
	self.frame.happiness = CreateFrame( "Button", "$parent_Happiness", self.frame );

	self.frame.happiness:SetScript( "OnClick", 
		function() 
			GameTooltip:Hide();
			nUI.PetFeeder:FeedPet( self.frame.happiness ); 
		end 
	);
	self.frame.happiness:SetFrameLevel( 3 );
	self.frame.happiness:SetBackdrop(
		{
			bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
			edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
			tile     = true, 
			tileSize = 1, 
			edgeSize = 10, 
			insets   = {left = 2, right = 2, top = 2, bottom = 2},
		}
	);
	self.frame.happiness:SetBackdropColor( 0, 0, 0, 0.75 );
	self.frame.happiness:SetWidth( dL + 2 );
	self.frame.happiness:SetHeight( dL + 2 );
	self.frame.happiness:SetFrameLevel( 5 );
	self.frame.happiness:SetAlpha( 0 );
	self.frame.happiness:EnableMouse( false );
	self.frame.happiness:SetScript( "OnLeave", nil );
	self.frame.happiness:SetScript( "OnEnter", nil );

	self.frame.happiness.texture = self.frame.happiness:CreateTexture( "$parentTexture" );
	self.frame.happiness.texture:SetPoint( "TOPLEFT", self.frame.happiness, "TOPLEFT", 3, -3 );
	self.frame.happiness.texture:SetPoint( "TOPRIGHT", self.frame.happiness, "TOPRIGHT", -3, -3 );
	self.frame.happiness.texture:SetPoint( "BOTTOMLEFT", self.frame.happiness, "BOTTOMLEFT", 3, 3 );
	self.frame.happiness.texture:SetPoint( "BOTTOMRIGHT", self.frame.happiness, "BOTTOMRIGHT", -3, 3 );
	self.frame.happiness.texture:SetTexture( "Interface\\PetPaperDollFrame\\UI-PetHappiness.blp" );
	
	if self.has_model then				
		if layout.orientation ~= "right" then
			self.frame.happiness:SetPoint( "TOPRIGHT", self.frame.model, "TOPRIGHT", 0, 0 );
		else
			self.frame.happiness:SetPoint( "TOPLEFT", self.frame.model, "TOPLEFT", 0, 0 );
		end
	else
		if layout.orientation ~= "right" then
			self.frame.happiness:SetPoint( "LEFT", self.frame.label, "RIGHT", 0, 0 );
		else
			self.frame.happiness:SetPoint( "RIGHT", self.frame.label, "LEFT", 0, 0 );
		end
	end
	
	-- create the status bar box if it is to be displayed
	
	if self.compact or layout.StatusBars.visible == "yes" then

		self.frame.status = self:createUnitFrame( "$parent_Status", self.frame, true );

		self.frame.status:SetFrameLevel( 3 );
		self.frame.status:SetBackdrop(
			{
				bgFile   = "Interface\\Tooltips\\UI-Tooltip-Background", 
				edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", 
				tile     = true, 
				tileSize = 1, 
				edgeSize = 10, 
				insets   = {left = 2, right = 2, top = 2, bottom = 2},
			}
		);
		self.frame.status:SetBackdropColor( 0, 0, 0, 0.75 );
		self.frame.status:SetWidth( dX+1 );
		
		if self.compact then
			self.frame.status:SetHeight( info_box_size );
		elseif layout.Size.two_line == "yes" and layout.show_detail ~= "no" then
			self.frame.status:SetHeight( info_box_size );
		else
			self.frame.status:SetHeight( info_box_size * 2 );
		end

		if self.compact then
			self.frame.status:SetPoint( "BOTTOM", self.frame.label, "BOTTOM", 0, 1 );
		elseif layout.orientation == "left" then
			self.frame.status:SetPoint( "TOPRIGHT", self.frame.label, "BOTTOMRIGHT", 0, 1 );
		else
			self.frame.status:SetPoint( "TOPLEFT", self.frame.label, "BOTTOMLEFT", 0, 1 );
		end
		
		-- calculate how we're fitting bars inside the box
		
		local	Hx   = dX - 4;
		local	Hy   = info_box_size - 4;
		local	Py   = 0;		-- height of the power bar
		local	bars = 1;		-- the nunmber of bars to display health/power/casting
		
		if layout.StatusBars.power ~= "none" then
			bars = bars+1;
		end
		
		if not self.compact and (layout.Size.two_line ~= "yes" or layout.show_detail == "no") then
			Hy = Hy + info_box_size;
		end
		
		if layout.StatusBars.power == "thin" then
			Py = math.floor( Hy / bars * 0.5 );
		elseif layout.StatusBars.power ~= "none" then
			Py = math.floor( Hy / bars );
		end
		
		Hy = Hy - Py;

		-- create the status bars as required
		
		self.frame.status.power   = nil;
		self.frame.status.latency = nil;

		layout.HealthBar.width   = Hx;
		layout.HealthBar.height  = Hy;
		self.frame.status.health = nUI.Class.UnitBar_Health:new( self.frame.status, self.id+self.rid, layout.HealthBar, unit_id );
		self.rid                 = self.rid + (self.frame.status.health.rid or 1);

		local	hook = layout.HealthBar.name;
		
		self.frame.status.health:SetPoint( "TOPLEFT", self.frame.status, "TOPLEFT", 2, -3 );

		if layout.StatusBars.power ~= "none" then
			
			layout.ManaBar.width     = Hx;
			layout.ManaBar.height    = Py;
			self.frame.status.mana   = nUI.Class.UnitBar_Mana:new( self.frame.status, self.id+self.rid, layout.ManaBar, unit_id );
			self.rid                 = self.rid + (self.frame.status.mana.rid or 1);
			
			layout.RageBar.width     = Hx;
			layout.RageBar.height    = Py;
			self.frame.status.rage   = nUI.Class.UnitBar_Rage:new( self.frame.status, self.id+self.rid, layout.RageBar, unit_id );
			self.rid                 = self.rid + (self.frame.status.rage.rid or 1);
			
			layout.EnergyBar.width   = Hx;
			layout.EnergyBar.height  = Py;
			self.frame.status.energy = nUI.Class.UnitBar_Energy:new( self.frame.status, self.id+self.rid, layout.EnergyBar, unit_id );
			self.rid                 = self.rid + (self.frame.status.energy.rid or 1);

			layout.FocusBar.width    = Hx;
			layout.FocusBar.height   = Py;
			self.frame.status.focus  = nUI.Class.UnitBar_Focus:new( self.frame.status, self.id+self.rid, layout.FocusBar, unit_id );
			self.rid                 = self.rid + (self.frame.status.focus.rid or 1);
			
			self.frame.status.mana:SetPoint( "TOPLEFT", hook, "BOTTOMLEFT", 0, 0 );
			self.frame.status.rage:SetPoint( "TOPLEFT", hook, "BOTTOMLEFT", 0, 0 );
			self.frame.status.focus:SetPoint( "TOPLEFT", hook, "BOTTOMLEFT", 0, 0 );
			self.frame.status.energy:SetPoint( "TOPLEFT", hook, "BOTTOMLEFT", 0, 0 );
			
			self.frame.status.mana:Hide();
			self.frame.status.rage:Hide();
			self.frame.status.focus:Hide();
			self.frame.status.energy:Hide();

			hook = layout.ManaBar.name;
			
		end
	end

	-- create a state overlay frame used to indicate AFK / Dead / etc.
	
	self.frame.state = CreateFrame( "Frame", "$parent_State", self.frame );
	self.frame.state:SetBackdrop(
		{
			bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
			edgeFile = nil, 
			tile = true, 
			tileSize = 1, 
			edgeSize = 10, 
			insets = {left = 2, right = 2, top = 2, bottom = 2},
		}
	);
	self.frame.state:SetFrameLevel( 8 );
	self.frame.state:SetBackdropColor( 0.5, 0.5, 0.5, 0.75 );
	self.frame.state:SetWidth( layout.Size.width * nUI_scale - 4 );
	self.frame.state:SetHeight( layout.Size.height * nUI_scale - 4 );
	self.frame.state:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", 2, -2 );

	self.frame.state.text = self.frame.state:CreateFontString( "$parent_StateText", "OVERLAY" );
	self.frame.state.text:SetFont( "Fonts\\ARIALN.TTF", dL * 0.7, "OUTLINE" );
	self.frame.state.text:SetTextColor( 1, 0.25, 0.25, 1 );
	self.frame.state.text:SetJustifyH( "CENTER" );
	self.frame.state.text:SetJustifyV( "MIDDLE" );
	
	if layout.orientation == "left" then
		self.frame.state.text:SetPoint( "RIGHT", self.frame.label, "RIGHT", -5, -dL * 0.025 );
	else
		self.frame.state.text:SetPoint( "LEFT", self.frame.label, "LEFT", 5,  -dL * 0.025 );
	end

	self.frame.state:SetAlpha( 0 );
			
	-- layout the unit buffs
	
	self.frame.buffs = nUI.Class.Buffs:new( 
		self.frame, self.id, 
		self:createBuffLayout( layout.BuffInfo.Buffs, "Buffs" ), 
		unit_id 
	);

	self:setBuffLocation( self.frame.buffs, layout.BuffInfo.Buffs );
	
	self.frame.debuffs = nUI.Class.Buffs:new( 
		self.frame, self.id, 
		self:createBuffLayout( layout.BuffInfo.Debuffs, "Debuffs" ), 
		unit_id 
	);

	self:setBuffLocation( self.frame.debuffs, layout.BuffInfo.Debuffs );
	
	self:setHighlightNormal();
	self:registerEvents();

end

-------------------------------------------------------------------------------

function prototype:createUnitFrame( name, parent, clickable )
	
	local frame   = CreateFrame( "Button", name, parent, "SecureUnitButtonTemplate" );
	
	self:initUnitFrame( frame, clickable );
	
	return frame;
end

-------------------------------------------------------------------------------

function prototype:initUnitFrame( frame, clickable )
	
	local id      = self.layout.id;
	local unit_id = self.layout.unit_id;
	
	frame:SetAttribute( "unit", unit_id );
	frame:SetAttribute( "*type1", "target" );
	frame:SetAttribute( "type2",  "menu" );
	
	frame.unit = unit_id;
	frame.id   = id;

	if id then frame:SetID( id ); end;
	
	if clickable then

		local click_cast = nUI_Config.click_cast ~= "no";
		
		frame:RegisterForClicks( "LeftButtonUp", "RightButtonUp" );	
		frame:SetScript( "OnLeave", function() GameTooltip:Hide(); end );
		
		frame:SetScript( "OnEnter", 
			function() 
				GameTooltip_SetDefaultAnchor( GameTooltip, frame:GetParent() );
				GameTooltip:SetUnit( self.layout.unit_id ); 
				
				if self.role then
					GameTooltip:AddLine( "Role: "..self.role, 1, 1, 1, 1 );
				end
				
				if self.unit_spec then 
					GameTooltip:AddLine( self.unit_spec, 1, 1, 1, 1 ); 
				end 
				
				GameTooltip:Show();
			end 
		);
	
		frame.menu = function()

			this.unit   = unit_id;
			this.id     = id;
			this.name,
			this.server = UnitName( unit_id );
						
			ToggleDropDownMenu( 1, nil, self.frame.popup, self.event_frame, 106, 27 );
		end
				
		ClickCastFrames[frame]     = click_cast;
		nUI.ClickCastFrames[frame] = true;
		
		table.insert( self.ClickCastFrames, frame );
		
	end
end

-------------------------------------------------------------------------------

function prototype:createBuffLayout( layout, type )
	
	local buff_layout;
	local visible = "yes";
	local model_side = self.layout.orientation;
	local cols       = math.ceil( 40 / layout.rows );
	local size
	local orient;
	local grow;
	
	if layout.location == "hidden" then visible = "hidden";
	elseif layout.location == "none" then visible = "no";
	end
	 
	if layout.location == "detatched" then
		orient = layout.anchor;
		grow   = "horizontal";
		size   = (layout.width or self.layout.Size.width) / cols;
	elseif layout.location == "left" then
		orient = "TOPRIGHT";
		grow   = "vertical";
		size   = self.layout.Size.height / cols;
	elseif layout.location == "right" then
		orient = "TOPLEFT";
		grow   = "vertical";
		size   = self.layout.Size.height / cols;
	elseif layout.location == "top" then
		orient = (model_side == "left" and "BOTTOMLEFT" or "BOTTOMRIGHT");
		grow   = "horizontal";
		size   = self.layout.Size.width / cols;
	else
		orient = (model_side == "left" and "TOPLEFT" or "TOPRIGHT");
		grow   = "horizontal";
		size   = self.layout.Size.width / cols;
	end
		
	buff_layout =
	{
		name       = self.frame_name.."_"..type,
		visible    = visible,
		flash      = layout.flash,
		flash_rate = layout.flash_rate,
		timers     = layout.timers,
		anim       = layout.anim,
		highlight  = layout.highlight,
		type       = layout.type,
		
		Shape =
		{
			orient = orient,
			grow   = grow,
			rows   = layout.rows,
			size   = (size - layout.gap) * nUI.scale,
			gap    = layout.gap * nUI.scale,
		},
	};
	
	if layout.timers then buff_layout.Shape.size = buff_layout.Shape.size * 4 / 3; end
	
	return nUI.Class.BuffsLayout:new( buff_layout );
		
end

-------------------------------------------------------------------------------

function prototype:setBuffLocation( frame, layout )
	
	local location = layout.location;
	local side     = self.layout.orientation;
	
	if location == "detatched" then
		
		frame:SetPoint( layout.anchor, nUI_Dashboard, "CENTER", layout.xOfs * nUI.scale, layout.yOfs * nUI.scale );
		
	elseif location == "left" then
		
		frame:SetPoint( "TOPRIGHT", self.frame, "TOPLEFT", 0, -3 );
		
	elseif location == "right" then
		
		frame:SetPoint( "TOPLEFT", self.frame, "TOPRIGHT", 0, -3 );
		
	elseif location == "top" then
		
		frame:SetPoint( "BOTTOMLEFT", self.frame, "TOPLEFT", (side == "left" and 3 or -3), 0 );
		
	else
		
		frame:SetPoint( "TOPLEFT", self.frame, "BOTTOMLEFT", (side == "left" and 3 or -3), 0 );
		
	end
	
end

-------------------------------------------------------------------------------

function prototype:SetPoint( point, relative_to, relative_point, ofsX, ofsY )
	
	self.frame:SetPoint( point, relative_to, relative_point, ofsX, ofsY );
	
end

function prototype:SetAllPoints( relative_to )
	
	self.frame:SetAllPoints( relative_to );
	
end

function prototype:ClearAllPoints()
	
	self.frame:ClearAllPoints();

end

-------------------------------------------------------------------------------

function prototype:SetBackdropColor( r, g, b, a )

	if self.has_model then self.frame.model:SetBackdropColor( r, g, b, a ); end;
	if self.frame.level then self.frame.level:SetBackdropColor( r, g, b, a ); end;
	if self.frame.level and self.frame.level.boss then self.frame.level.boss:SetBackdropColor( r, g, b, a ); end;
	if self.frame.unit_class then self.frame.unit_class:SetBackdropColor( r, g, b, a ); end;
	if self.frame.status then self.frame.status:SetBackdropColor( r, g, b, a ); end;
	
end

function prototype:SetBackdropBorderColor( r, g, b, a )
	
	if self.has_model then self.frame.model:SetBackdropBorderColor( r, g, b, a ); end;
	if self.frame.level then self.frame.level:SetBackdropBorderColor( r, g, b, a ); end;
	if self.frame.unit_class then self.frame.unit_class:SetBackdropBorderColor( r, g, b, a ); end;
	if self.frame.status then self.frame.status:SetBackdropBorderColor( r, g, b, a ); end;
	
	self.frame.label:SetBackdropBorderColor( r, g, b, a ); 

end

-------------------------------------------------------------------------------

function prototype:setHighlightAgro()
	
	self:SetBackdropColor( 0.5, 0, 0, 1 );
	self:SetBackdropBorderColor( 1, 0.3, 0.3, 0.75 );
	
end

function prototype:setHighlightHeal()
	
	self:SetBackdropColor( 0, 0.75, 0, 1 );
	self:SetBackdropBorderColor( 0.2, 1, 0.2, 0.75 );
	
end

function prototype:setHighlightCurse()
	
	self:SetBackdropColor( 0, 0.5, 0.5, 1 );
	self:SetBackdropBorderColor( .25, 1, 1, 0.75 );
	
end

function prototype:setHighlightNormal()
	
	self:SetBackdropColor( 0, 0, 0, 0.75 );
	self:SetBackdropBorderColor( 1, 1, 1, 0.75 );
	
end

-------------------------------------------------------------------------------

function prototype:enableAnim( mode )

	if mode == "no" then self.layout.Portrait.visible = "yes";
	else self.layout.Portrait.visible = "anim";
	end

	self:updateModel();
	
end

-------------------------------------------------------------------------------

function prototype:updateRaidInfo()
	
	self.RaidInfo = {};
	self.role     = nil;

	local frame            = self.frame;
	local num_raid_members = GetNumRaidMembers();
	
	self.RaidInfo.in_raid  = num_raid_members > 0 and UnitInRaid( self.layout.unit_id );
	self.RaidInfo.in_party = GetNumPartyMembers() > 0 and UnitInParty( self.layout.unit_id );
	
	if self.RaidInfo.in_raid then
			
		local raid_index = self.layout.raid_id;
		
		for i=(raid_index or 1),(raid_index or num_raid_members) do
			
			self.RaidInfo.name,self.RaidInfo.rank,self.RaidInfo.subGroup,self.RaidInfo.level,
			self.RaidInfo.class,self.RaidInfo.fileName,self.RaidInfo.zone,self.RaidInfo.online,
			self.RaidInfo.isDead,self.RaidInfo.role,self.RaidInfo.is_ml = GetRaidRosterInfo( i );
			
			if not raid_index 
			and self.RaidInfo.name
			and UnitIsUnit( self.layout.unit_id, self.RaidInfo.name ) then 
				break; 
			end
		end
	end
	
	-- this code checks to see if the unit is a party leader or not. Bliz's own code
	-- does not rely on the party leader change event and even notes that event to
	-- be buggy, so we follow their lead and check in the update routine.

	if not self.RaidInfo.in_raid and not self.RaidInfo.in_party then
		
		local raid_info = self.RaidInfo;
		
		if not raid_info.is_solo then
			
			raid_info.is_solo = true;
			
			if self.is_leader or self.is_assist then
				
				self.is_leader = false;
				self.is_assist = false;
				frame.leader:SetAlpha( 0 );
				
			end
				
			if self.main_tank or self.off_tank then
				
				self.main_tank = false;
				self.off_tank  = false;
				frame.tank:SetAlpha( 0 );
				
			end
			
			frame.master_looter:SetAlpha( 0 );
			
			if self.frame.group and self.frame.group.group_id then
				self.frame.group.group_id = nil;
				self.frame.group:SetText( "" );
				self.frame.raid:SetAlpha( 0 );
			end
		end
		
	else
		
		local raid_info = self.RaidInfo;
		
		raid_info.is_solo = false;
		
		if (raid_info.in_raid and raid_info.rank == 2) 
		or (raid_info.in_party and UnitIsPartyLeader( unit_id )) then
			
			if raid_info.in_raid then
				self.role = (self.role and "/"..self.role or "").."Raid Leader";
			else
				self.role = (self.rolw and "/"..self.role or "").."Party Leader";
			end
			
			if not self.is_leader then
				
				frame.leader.texture:SetTexture( "Interface\\GroupFrame\\UI-Group-LeaderIcon.blp" );
				frame.leader.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
				frame.leader:SetAlpha( 1 );
				self.is_leader = true;
				self.is_assist = false;

			end
			
		elseif raid_info.in_raid and raid_info.rank == 1 then
			
			self.role = (self.role and "/"..self.role or "").."Raid Assist";
			
			if not self.is_assist then
				
				frame.leader.texture:SetTexture( "Interface\\GroupFrame\\UI-GROUP-ASSISTANTICON.blp" );
				frame.leader.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
				frame.leader:SetAlpha( 1 );
				self.is_leader = false;
				self.is_assist = true;

			end
			
		elseif self.is_leader or self.is_assist then
			
			self.frame.leader:SetAlpha( 0 );
			self.is_leader = false;
			self.is_assist = false;
			
		end
		
		-- display the raid group we are in
		
		if self.frame.group then
			
			if not raid_info.in_raid then
				
				if self.frame.group.group_id then
			
					self.frame.group.group_id = nil;
					self.frame.group:SetText( "" );
					self.frame.raid:SetAlpha( 0 );
					
				end
			
			else

				if self.frame.group.group_id ~= raid_info.subGroup then
					
					self.frame.group.group_id = raid_info.subGroup;
					self.frame.group:SetText( raid_info.subGroup );
					self.frame.raid:SetAlpha( 1 );
					
				end
			
			end
		end
		
		-- if we are in a raid, check to see if we are master looter, main tank and/or off-tank
		
		if raid_info.in_raid then
			
			if raid_info.is_ml and not frame.master_looter:GetAlpha() ~= 1 then
				self.role = (self.role and "/"..self.role or "").."Master Looter";
				frame.master_looter:SetAlpha( 1 );
			elseif not raid_info.is_ml and frame.master_looter:GetAlpha() ~= 0 then
				frame.master_looter:SetAlpha( 0 );
			end
			
			if not self.main_tank and raid_info.role == "maintank" then
				self.role = (self.role and "/"..self.role or "").."Main Tank";
				frame.tank.texture:SetTexture( "Interface\\GroupFrame\\UI-GROUP-MAINTANKICON.blp" );
				frame.tank.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
				frame.tank:SetAlpha( 1 );
				self.main_tank = true;
				self.off_tank  = false;
			elseif not self.off_tank and raid_info.role == "mainassist" then
				self.role = (self.role and "/"..self.role or "").."Off-Tank";
				frame.tank.texture:SetTexture( "Interface\\GroupFrame\\UI-GROUP-MAINASSISTICON.blp" );
				frame.tank.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
				frame.tank:SetAlpha( 1 );
				self.main_tank = false;
				self.off_tank  = true;
			elseif self.main_tank or self.off_tank then
				frame.tank:SetAlpha( 0 );
				self.main_tank = false;
				self.off_tank  = false;
			end
		end
	end	
end

-------------------------------------------------------------------------------
--                               EVENT HANDLING                              --
-------------------------------------------------------------------------------

function prototype:registerEvents( is_init )

	-- callbacks
	
	self.frame:SetScript( "OnEvent", function() self:onEvent(); end );
	self.frame:SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elapsed ); end );
--	self.frame:SetScript( "OnShow", function() DEFAULT_CHAT_FRAME:AddMessage( self.layout.unit_id.." is visible", 1, 0.83, 0 ); self:checkGUID(); end );
--	self.frame:SetScript( "OnHide", function() DEFAULT_CHAT_FRAME:AddMessage( self.layout.unit_id.." is hidden", 1, 0.83, 0 ); self.guid = nil; end );
	self.frame:SetScript( "OnShow", function() self:checkGUID(); end );
	self.frame:SetScript( "OnHide", function() self.guid = nil; end );

	-- events
	
	self.frame:RegisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame:RegisterEvent( "UNIT_LEVEL" );
	self.frame:RegisterEvent( "UNIT_MODEL_CHANGED" );
	self.frame:RegisterEvent( "UNIT_NAME_UPDATE" );
	self.frame:RegisterEvent( "UNIT_PORTRAIT_UPDATE" );
	self.frame:RegisterEvent( "UNIT_COMBAT" );	
	self.frame:RegisterEvent( "UNIT_HAPPINESS" );
	self.frame:RegisterEvent( "UNIT_MAXHAPPINESS" );
	self.frame:RegisterEvent( "UNIT_FACTION" );
	self.frame:RegisterEvent( "UNIT_DYNAMIC_FLAGS" );
	self.frame:RegisterEvent( "UNIT_FLAGS" );
	self.frame:RegisterEvent( "UNIT_TARGET" );
	self.frame:RegisterEvent( "PARTY_MEMBERS_CHANGED" );	
	self.frame:RegisterEvent( "RAID_ROSTER_UPDATE" );	
	self.frame:RegisterEvent( "RAID_TARGET_UPDATE" );	
	self.frame:RegisterEvent( "UNIT_PET" );	
	self.frame:RegisterEvent( "PET_UI_UPDATE" );	
	self.frame:RegisterEvent( "PLAYER_FLAGS_CHANGED" );	
	self.frame:RegisterEvent( "PLAYER_ALIVE" );	
	self.frame:RegisterEvent( "PLAYER_DEAD" );	
	self.frame:RegisterEvent( "PLAYER_UNGHOST" );	
	self.frame:RegisterEvent( "PLAYER_TARGET_CHANGED" );	
	self.frame:RegisterEvent( "PLAYER_FOCUS_CHANGED" );	
	self.frame:RegisterEvent( "PLAYER_UPDATE_RESTING" );	
	self.frame:RegisterEvent( "PLAYER_ENTER_COMBAT" );	
	self.frame:RegisterEvent( "PLAYER_LEAVE_COMBAT" );	
	self.frame:RegisterEvent( "PLAYER_REGEN_DISABLED" );	
	self.frame:RegisterEvent( "PLAYER_REGEN_ENABLED" );	
	self.frame:RegisterEvent( "PET_ATTACK_START" );	
	self.frame:RegisterEvent( "PET_ATTACK_STOP" );	

	-- set up combat feedback on those units that have models

	if self.has_model then
		
		self.CombatFeedback = function( arg2, arg3, arg4, arg5 )
--			self.frame.model:combat_event( arg2, arg3, arg4, arg5 );
		end
				
		local initFeedback = function()
			
			this = self.frame.model;

--			CombatFeedback_Initialize( self.dY * 0.29 );
		end;
		
		initFeedback();
		
		self.frame.model.feedbackStartTime = 0;		
--		self.frame.model:SetScript( "OnUpdate", CombatFeedback_OnUpdate );
				
	end		
	
	if not is_init then
		
		if self.frame.status then
			self.frame.status.health:registerEvents();
			self.frame.status.mana:registerEvents();
			self.frame.status.rage:registerEvents();
			self.frame.status.energy:registerEvents();
			self.frame.status.focus:registerEvents();
		end
		
		if self.frame.casting then self.frame.casting:registerEvents(); end
		if self.frame.channel then self.frame.channel:registerEvents(); end
		if self.frame.latency then self.frame.latency:registerEvents(); end
		if self.frame.buffs then self.frame.buffs:registerEvents(); end
		if self.frame.debuffs then self.frame.debuffs:registerEvents(); end
		
	end	
end

-------------------------------------------------------------------------------

function prototype:unregisterEvents()

	-- callbacks
	
	self.frame:SetScript( "OnEvent", nil );
	self.frame:SetScript( "OnUpdate", nil );
	self.frame:SetScript( "OnShow", nil );
	self.frame:SetScript( "OnHide", nil );

	-- events
	
	self.frame:UnregisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame:UnregisterEvent( "UNIT_LEVEL" );
	self.frame:UnregisterEvent( "UNIT_MODEL_CHANGED" );
	self.frame:UnregisterEvent( "UNIT_NAME_UPDATE" );
	self.frame:UnregisterEvent( "UNIT_PORTRAIT_UPDATE" );
	self.frame:UnregisterEvent( "UNIT_COMBAT" );	
	self.frame:UnregisterEvent( "UNIT_HAPPINESS" );
	self.frame:UnregisterEvent( "UNIT_MAXHAPPINESS" );
	self.frame:UnregisterEvent( "UNIT_FACTION" );
	self.frame:UnregisterEvent( "UNIT_DYNAMIC_FLAGS" );
	self.frame:UnregisterEvent( "UNIT_FLAGS" );
	self.frame:UnregisterEvent( "UNIT_TARGET" );
	self.frame:UnregisterEvent( "PARTY_MEMBERS_CHANGED" );	
	self.frame:UnregisterEvent( "RAID_ROSTER_UPDATE" );	
	self.frame:UnregisterEvent( "RAID_TARGET_UPDATE" );	
	self.frame:UnregisterEvent( "UNIT_PET" );	
	self.frame:UnregisterEvent( "PET_UI_UPDATE" );	
	self.frame:UnregisterEvent( "PLAYER_FLAGS_CHANGED" );	
	self.frame:UnregisterEvent( "PLAYER_ALIVE" );	
	self.frame:UnregisterEvent( "PLAYER_DEAD" );	
	self.frame:UnregisterEvent( "PLAYER_UNGHOST" );	
	self.frame:UnregisterEvent( "PLAYER_TARGET_CHANGED" );	
	self.frame:UnregisterEvent( "PLAYER_FOCUS_CHANGED" );	
	self.frame:UnregisterEvent( "PLAYER_UPDATE_RESTING" );	
	self.frame:UnregisterEvent( "PLAYER_ENTER_COMBAT" );	
	self.frame:UnregisterEvent( "PLAYER_LEAVE_COMBAT" );	
	self.frame:UnregisterEvent( "PLAYER_REGEN_DISABLED" );	
	self.frame:UnregisterEvent( "PLAYER_REGEN_ENABLED" );	
	self.frame:UnregisterEvent( "PET_ATTACK_START" );	
	self.frame:UnregisterEvent( "PET_ATTACK_STOP" );	

	if self.has_model then		
		self.frame.model:SetScript( "OnUpdate", nil );				
	end		
	
	if self.frame.status then
		self.frame.status.health:unregisterEvents();
		self.frame.status.mana:unregisterEvents();
		self.frame.status.rage:unregisterEvents();
		self.frame.status.energy:unregisterEvents();
		self.frame.status.focus:unregisterEvents();
	end
	
	if self.frame.casting then self.frame.casting:unregisterEvents(); end
	if self.frame.channel then self.frame.channel:unregisterEvents(); end
	if self.frame.latency then self.frame.latency:unregisterEvents(); end
	if self.frame.buffs then self.frame.buffs:unregisterEvents(); end
	if self.frame.debuffs then self.frame.debuffs:unregisterEvents(); end
end

-------------------------------------------------------------------------------

function prototype:onEvent()
	
	local unit_id = self.layout.unit_id;
	
	-- unit specific events...

	if     event == "PARTY_MEMBERS_CHANGED" then self:checkGUID(); self:updateRaidInfo();
	elseif event == "RAID_ROSTER_UPDATE"    then self:checkGUID(); self:updateRaidInfo();
	elseif event == "PET_UI_UPDATE"         then self:checkGUID();
	elseif event == "PLAYER_TARGET_CHANGED" then self:checkGUID();
	elseif event == "UNIT_TARGET"           then self:checkGUID();
	elseif event == "PLAYER_FOCUS_CHANGED"  then self:checkGUID();
	elseif event == "RAID_TARGET_UPDATE"    then self:updateRaidTarget();
	elseif event == "PLAYER_UPDATE_RESTING" then self:updateResting();
	elseif event == "UNIT_NAME_UPDATE"      then self:checkGUID();

	elseif event == "UNIT_PET" and unit_id:match( ".pet." ) then
	
		self:checkGUID();

	elseif UnitIsUnit( unit_id, (arg1 or unit_id) ) then
		if     event == "UNIT_DISPLAYPOWER"     then self:updatePowerBar(); 
		elseif event == "UNIT_LEVEL"            then self:updateUnitLevel(); 
		elseif event == "UNIT_HAPPINESS"        then self:updateHappiness(); 
		elseif event == "UNIT_MAXHAPPINESS"     then self:updateHappiness(); 
		elseif event == "UNIT_PORTRAIT_UPDATE"  then self:updateModel();
		elseif event == "UNIT_FACTION"          then self:updateFaction();
		elseif event == "UNIT_COMBAT"           then self:updateCombatStatus( arg2, arg3, arg4, arg5 );
		elseif event == "UNIT_MODELCHANGED"     then self:updateModel();
		elseif event == "UNIT_DYNAMIC_FLAGS"    then self:updateStatus( arg1 );
		elseif event == "UNIT_FLAGS"            then self:updateStatus( arg1 );
		elseif event == "PLAYER_FLAGS_CHANGED"  then self:updateStatus( arg1 ); self:updateRaidInfo();
		
		elseif UnitIsUnit( "player", unit_id ) then
			if     event == "PLAYER_UPDATE_RESTING" then self:updateResting();
			elseif event == "PLAYER_ENTER_COMBAT"	then self:enterCombat();
			elseif event == "PLAYER_REGEN_DISABLED" then self:enterCombat();
			elseif event == "PLAYER_LEAVE_COMBAT"   then self:leaveCombat();
			elseif event == "PLAYER_REGEN_ENABLED"  then self:leaveCombat();
			elseif event == "PLAYER_ALIVE"          then self:updateStatus( arg1 );
			elseif event == "PLAYER_DEAD"           then self:updateStatus( arg1 );
			elseif event == "PLAYER_UNGHOST"        then self:updateStatus( arg1 );
			end		
		
		elseif UnitIsUnit( "pet", unit_id ) then
			
			if event     == "PET_ATTACK_START" then self:enterCombat();
			elseif event == "PET_ATTACK_STOP" then self:leaveCombat();
			end
		end
	end
end

-------------------------------------------------------------------------------

function prototype:onUpdate( elapsed )
	
	if self.parent:IsVisible() then
		
		self.Timers.refresh = self.Timers.refresh + elapsed;
	
		-- update unit frame visibility at 4fps

		if self.Timers.refresh > 0.25 then
	
			self.Timers.refresh = 0;		
			self:refreshUnit();
			self:updateStatus( self.layout.unit_id );
		end
	end		
end

-------------------------------------------------------------------------------

function prototype:updateUnitPopup()

	local unit_id    = self.layout.unit_id;
	local unit_popup = nil;
	local raid_id    = nil;
	local unit_name  = nil;
	
	-- select the popup menu to be displayed
	
	if UnitIsUnit( "player", unit_id ) then
		
		unit_popup = "SELF";
		
	elseif UnitIsUnit( "pet", unit_id ) then
		
		unit_popup = "PET";

	elseif UnitIsPlayer( unit_id ) then
		
		raid_id = UnitInRaid( unit_id );
		
		if raid_id then
			
			unit_popup = "RAID_PLAYER";
			unit_name  = self.unit_name;
			
		elseif UnitInParty( unit_id ) then
			
			unit_popup = "PARTY";
			
		else
			
			unit_popup = "PLAYER";
			
		end
	else
		unit_popup = "RAID_TARGET_ICON";
		unit_name  = RAID_TARGET_ICON;
	end
	
	-- if we have one, then enable it
	
	if unit_popup then
		
		UnitPopup_ShowMenu( self.frame.popup, unit_popup, unit_id, unit_name, raid_id );
	
	end

end

-------------------------------------------------------------------------------

function prototype:updateUnitLevel()

	local	is_a_ding = false;
	local	new_level = UnitLevel( self.layout.unit_id );

	if self.unit_level ~= nil and self.unit_level ~= new_level and self.layout.unit_id == "player" then
		is_a_ding = true;
	end
	
	self.unit_level = new_level;
	
	if self.frame.level then
		
		self.frame.level.text:SetText( self.unit_level ); 

		-- if the unit is too high level, display the boss icon instead of the level
		
		if self.frame.level then
			if self.unit_level == -1 then	self.frame.level.boss:SetAlpha( 1 );
			else							self.frame.level.boss:SetAlpha( 0 );
			end
		end
		
		-- color the level text according to the differnce between the player and the unit
		-- based also on the unit's reaction to the player
				
		local level_dif = UnitLevel( "player" ) - self.unit_level;

		if self.unit_level == -1 then level_dif = -5; end
								
		if self.is_friend then			self.frame.level.text:SetTextColor( 1, 0.8196079, 0, 1 );
		elseif level_dif >= 5 then		self.frame.level.text:SetTextColor( 0.65, 0.65, 0.65, 1 );
		elseif level_dif >= 3 then		self.frame.level.text:SetTextColor( 0, 1, 0, 1 );
		elseif level_dif >= -2 then		self.frame.level.text:SetTextColor( 1, 0.8196079, 0, 1 );
		else							self.frame.level.text:SetTextColor( 1, 0.25, 0.25, 1 );
		end
		
		-- auto ding option
		
		if is_a_ding and nUI_Config.auto_ding == "yes" then
			nUI:autoDing( new_level );
		end
	end		
end
	
-------------------------------------------------------------------------------

function prototype:updatePowerBar()

	-- update health and power bars as required
	
	if self.frame.status then

		self.frame.status.health:Show();

		local power_type = UnitPowerType( self.layout.unit_id );
			
		if self.frame.status.power ~= nil then
			self.frame.status.power:Hide();
			self.frame.status.power = nil;
		end

		if power_type == 0 then
			self.frame.status.power = self.frame.status.mana;
		elseif power_type == 1 then
			self.frame.status.power = self.frame.status.rage;
		elseif power_type == 2 then
			self.frame.status.power = self.frame.status.focus;
		elseif power_type == 3 then
			self.frame.status.power = self.frame.status.energy;
		else
			nUI:debugMsg( 1, "unhandled unit power type ["..power_type.."] for unit ["..self.layout.unit_id.."]" );
		end

		if self.frame.status.power ~= nil then
			self.frame.status.power:Show();
		end
	end
end

-------------------------------------------------------------------------------

function prototype:updateHappiness()

	local unit_id = self.layout.unit_id;
	
	-- if this unit doesn't exist or is not the player's pet, hide the happiness indicator
	
	if not UnitExists( unit_id ) or not UnitIsUnit( "pet", unit_id ) then

		self.happiness = nil;
		self.loyalty   = nil;
		
		if self.frame.happiness then
			
			self.frame.happiness:SetAlpha( 0 );
			self.frame.happiness:EnableMouse( false );
			self.frame.happiness:RegisterForClicks();
			self.frame.happiness:SetScript( "OnLeave", nil );
			self.frame.happiness:SetScript( "OnEnter", nil );
				
		end
	
	-- otherwise... update status as appropriate
	
	else

		local happiness, _, loyalty = GetPetHappiness();
		local _, hunter_pet         = HasPetUI();

		-- this only applies to hunter pets
		
		if not happiness or not hunter_pet then
				
			self.happiness = nil;
			self.loyalty   = nil;

			if self.frame.happiness and self.frame.happiness:GetAlpha() ~= 0 then
				
				self.frame.happiness:SetAlpha( 0 );
				self.frame.happiness:EnableMouse( false );
				self.frame.happiness:RegisterForClicks();
				self.frame.happiness:SetScript( "OnLeave", nil );
				self.frame.happiness:SetScript( "OnEnter", nil );
			end

		-- this is the player's hunter pet...
		
		else
			
			local showFeedTooltip = function()
				
				GameTooltip:SetOwner( self.frame.happiness );
				if not nUI.PetFeeder:IsShown() then
					GameTooltip:SetText( "Click to feed "..UnitName( "pet" ) );
				else
					GameTooltip:SetText( "Click to cancel feeding" );
				end
				GameTooltip:Show();
				
			end
			
			-- make sure we can see the indicator
			
			if self.frame.happiness and self.frame.happiness:GetAlpha() ~= 1 then
				
				self.frame.happiness:SetAlpha( 1 );
				self.frame.happiness:EnableMouse( true );
				self.frame.happiness:RegisterForClicks( "LeftButtonUp" );
				self.frame.happiness:SetScript( "OnLeave", function() GameTooltip:Hide(); end );
				self.frame.happiness:SetScript( "OnEnter", function() showFeedTooltip(); end );
				
			end
			
			-- update the player if the pet's happiness status changes
			
			if happiness ~= self.happiness then

				-- if this is the player's pet unit frame, then announce the change in chat
				
				if unit_id == "pet" then

					if happiness == 1 then
						
						DEFAULT_CHAT_FRAME:AddMessage( "nUI: Warning! "..self.unit_name.." is |cFFFFFFFFNOT|r happy! Better feed soon.", 1, 0.5, 0.5 );
						
						nUI.PetFeeder:PrintPetFoods( 1, 0.5, 0.5 );
						
					elseif happiness == 2 then
						
						DEFAULT_CHAT_FRAME:AddMessage( "nUI: "..self.unit_name.." is unhappy... time to feed!", 1, 0.83, 0 );
						
						nUI.PetFeeder:PrintPetFoods( 1, 0.83, 0 );
						
					elseif happiness == 3 then
						
						DEFAULT_CHAT_FRAME:AddMessage( "nUI: "..self.unit_name.." is happy.", 0.5, 1, 0.5 );
						
					end
				end
				
				if self.frame.happiness then
									
					self.frame.happiness:SetAlpha( 1 );
					self.frame.happiness:EnableMouse( true );
					self.frame.happiness:RegisterForClicks( "LeftButtonUp" );
					self.frame.happiness:SetScript( "OnLeave", function() GameTooltip:Hide(); end );
					self.frame.happiness:SetScript( "OnEnter", function() showFeedTooltip(); end );
					
					if happiness == 1 then
						if unit_id == "pet" then
							self.frame.happiness:SetScale( 1.75 );
						end
						self.frame.happiness.texture:SetTexCoord(0.375, 0.5625, 0, 0.359375);
					elseif happiness == 2 then
						if unit_id == "pet" then
							self.frame.happiness:SetScale( 1.5 );
						end
						self.frame.happiness.texture:SetTexCoord(0.1875, 0.375, 0, 0.359375);
					elseif happiness == 3 then
						if unit_id == "pet" then
							self.frame.happiness:SetScale( 1 );
						end
						self.frame.happiness.texture:SetTexCoord(0, 0.1875, 0, 0.359375);
					end
				end
				
				self.happiness = happiness;
				
			end

			-- update the player if the pet's loyalty status changes
			
			if loyalty ~= self.loyalty then

				-- if this is the player's pet unit frame, then announce the change in chat
				
				if unit_id == "pet" then
					
					if loyalty < 0 then
						
						DEFAULT_CHAT_FRAME:AddMessage( "nUI: Warning... "..self.unit_name.." is "..(loyalty == -29 and "quickly" or "slowly").." losing loyalty ", 1, 0.5, 0.5 );
						
					elseif loyalty > 0 then
						
						DEFAULT_CHAT_FRAME:AddMessage( "nUI: "..self.unit_name.." is "..(loyalty == 20 and "quickly" or "slowly").." gaining loyalty", 0.5, 1, 0.5 );
						
					elseif self.loyalty < 0 then
						
						DEFAULT_CHAT_FRAME:AddMessage( "nUI: "..self.unit_name.." has stopped losing loyalty", 1, 0.83, 0 );
						
					elseif self.loyalty > 0 then
						
						DEFAULT_CHAT_FRAME:AddMessage( "nUI: "..self.unit_name.." has stopped gaining loyalty", 1, 0.83, 0 );
						
					end
				end
				
				self.loyalty = loyalty;
				
			end
		end		
	end
end

-------------------------------------------------------------------------------

function prototype:updateModel()

	local unit_id = self.layout.unit_id;
		
	if UnitExists( unit_id ) and self.has_model then
	
		-- if we need to swap in the unit portrait...
		
		if self.frame.model.portrait == nil and (self.layout.Portrait.visible == "yes" or UnitIsVisible( unit_id ) == nil) then

			if self.frame.model.anim then
				self.Cache.anim:SetAlpha( 0 );
				self.frame.model.anim = nil;
			end
	
			self.frame.model.portrait = self.Cache.portrait;
			self.frame.model.portrait:SetAlpha( 1 );

		-- if we need to swap in the unit animation...
		
		elseif self.frame.model.anim == nil and self.layout.Portrait.visible == "anim" and UnitIsVisible( unit_id ) then

			if self.frame.model.portrait then
				self.frame.model.portrait:SetAlpha( 0 );
				self.frame.model.portrait = nil;
			end

			self.frame.model.anim = self.Cache.anim;
			self.frame.model.anim:SetAlpha( 1 );
		end

		-- refresh the animation or portrait as the case may be...
		
		if self.frame.model.anim then 
			self.frame.model.anim:SetUnit( unit_id );
			self.frame.model.anim:SetCamera( 0 );
		elseif self.frame.model.portrait then
			SetPortraitTexture( self.frame.model.portrait.texture, unit_id ); 
		end	
	end					
end	

-------------------------------------------------------------------------------

function prototype:checkGUID()

	local unit_id = self.layout.unit_id;
	local exists  = UnitExists( unit_id );
	local visible = self.frame:IsShown();

	-- see if the object this unit refers to changed
	
	if exists then
	
		-- this is a globally unique identifier for every toon and mob in the world
		
		local guid = UnitGUID( unit_id );
		
		-- if the identifier changed, then we need to redefine the unit accordingly
		
		if guid ~= self.guid then
			
			HideDropDownMenu( 1 );
			
			-- remember who we're talking about now
			
			self.guid      = guid;
			self.unit_name,
			self.server    = UnitName( unit_id );

			-- this addresses a quirk when zoning that sometimes party member names
			-- are not immediately available. The same applies when first summoning a 
			-- pet from time to time. If we don't know which unit we are talking about,
			-- then we'll try again on the next update
			
			if not self.unit_name or self.unit_name == "Unknown" then
				
				self.guid = nil;
			
			-- otherwise, if we know who the new unit is, 
			-- update the unit content accordingly
			
			else
				
				self.frame.label.text:SetText( self.unit_name );
									
				self:updateUnitClass();
				self:updatePowerBar();
				self:updateModel();
				self:updateUnitInfo();
				self:updateFaction();
				self:updateHappiness();
				self:updateResting();
				self:updateRaidTarget();
				self:updateStatus( unit_id );	
				self:updateRaidInfo();
			end			
		end
	end
end

-------------------------------------------------------------------------------

function prototype:updateCombatStatus( arg2, arg3, arg4, arg5 )

	if self.CombatFeedback and UnitIsUnit( arg1, self.layout.unit_id ) then
		self.CombatFeedback( arg2, arg3, arg4, arg5 );
	end
end

-------------------------------------------------------------------------------

function prototype:updateFaction()

	local unit_id  = self.layout.unit_id;
	local pvp_icon = self.frame.pvp;
	local is_pvp   = UnitIsPVP( unit_id );
	local is_ffa   = UnitIsPVPFreeForAll( unit_id );
	local factionGroup, factionName = UnitFactionGroup( unit_id );

	-- update the unit's PVP status
	
	if is_ffa then
		
		if pvp_icon:GetAlpha() ~= 1 then

			if unit_id == "player" then
			
				PlaySound("igPVPUpdate");
	
				pvp_icon:SetScript( "OnEnter", 
					function()
						GameTooltip_AddNewbieTip( PVPFFA, 1.0, 1.0, 1.0, NEWBIE_TOOLTIP_PVPFFA, 1);
					end
				);
				pvp_icon:SetScript( "OnLeave", 
					function() 
						GameTooltip:Hide(); 
					end 
				);
			
			end
			
			pvp_icon.texture:SetTexture( "Interface\\TargetingFrame\\UI-PVP-FFA" );
			pvp_icon.texture:SetTexCoord( 0.03125, 0, 0.03125, 0.59375, 0.625, 0, 0.625, 0.59375 );
			pvp_icon:SetAlpha( 1 );
		end
		
	elseif factionGroup and is_pvp then
		
		if pvp_icon:GetAlpha() ~= 1 then
				
			if unit_id == "player" then
	
				PlaySound("igPVPUpdate");
				PlayerPVPIconHitArea.tooltipTitle = factionName;
				PlayerPVPIconHitArea.tooltipText = getglobal("NEWBIE_TOOLTIP_"..strupper(factionGroup));
				PlayerPVPIconHitArea:Show();
	
				pvp_icon:SetScript( "OnEnter", 
					function()
						GameTooltip_AddNewbieTip( factionName, 1.0, 1.0, 1.0, getglobal( "NEWBIE_TOOLTIP_"..strupper(factionGroup) ), 1);
					end
				);
				pvp_icon:SetScript( "OnLeave", 
					function() 
						GameTooltip:Hide(); 
					end 
				);
						
			end
			
			pvp_icon.texture:SetTexture("Interface\\TargetingFrame\\UI-PVP-"..factionGroup);
			pvp_icon.texture:SetTexCoord( 0.03125, 0, 0.03125, 0.59375, 0.625, 0, 0.625, 0.59375 );
			pvp_icon:SetAlpha( 1 );
		end
		
	elseif pvp_icon:GetAlpha() ~= 0 then
		
		pvp_icon:SetAlpha( 0 );
		pvp_icon:SetScript( "OnEnter", nil );
		pvp_icon:SetScript( "OnLeave", nil );

	end
	
	-- update the unit's faction
	
	local r, g, b, a = 0, 0, 0, 0.75;

	self.is_friend = true;

	-- if the unit is the player, the player's pet or a party or raid member, then
	-- there are only two possible states -- no reaction or colored green for PvP
	
	if UnitIsUnit( "player", unit_id ) or UnitIsUnit( "pet", unit_id ) or UnitPlayerOrPetInParty( unit_id ) == 1 or UnitPlayerOrPetInRaid( unit_id ) == 1 then
		
		if is_pvp or is_ffa then
			
			g = 0.5;
			
		end
		
	-- is this unit another player?
	
	elseif UnitPlayerControlled( unit_id ) then
			
		-- is the other player hostile to this player?
		
		if UnitCanAttack( unit_id, "player" ) then
			
			if UnitCanAttack( "player", unit_id ) then
				
				self.is_friend = false;
				r = 0.5;
			
			else
				
				b = 0.5;
				
			end
			
		-- could this player be hostile to the other player?
		
		elseif UnitCanAttack( "player", unit_id ) then
			
			r = 0.5;
			g = 0.5;
			self.is_friend = false;
		
		-- other players this player can assist are colored green
		
		elseif is_pvp and not UnitIsPVPSanctuary( "player" ) and not UnitIsPVPSanctuary( unit_id ) then
			
			g = 0.5;

		-- all other players are blue
		
		else
			
			b = 0.5;
			
		end

	-- could the player attack this unit?
	
	elseif UnitCanAttack( "player", unit_id ) then

		local reaction   = UnitReaction( "player", unit_id );
		local unit_level = UnitLevel( unit_id );
		
		self.if_friend = false;
		
		-- has someone else already tagged this mob?
		
		if UnitIsTapped( unit_id ) and not UnitIsTappedByPlayer( unit_id ) then
		
			r = 0.5;
			g = 0.5;
			b = 0.5;
			
		-- if it's hostile to the player, it's red
		
		elseif reaction >= 0 and reaction <= 3 then

			r = 0.5;
		
		-- otherwise it's yellow
		
		else				

			r = 0.5;
			g = 0.5;
			a = 0.90;
			
		end							
	
	
	-- paint friendly NPCs green
	
	elseif UnitIsFriend( "player", unit_id ) then

		g = 0.5;
			
	end

	-- paint the unit's label bar according to it's reaction to us
	
	if self.layout.show_reaction ~= "no" then
		self.frame.label:SetBackdropColor( r, g, b, a );
	end	
	
	-- update the unit level to ensure it is colored correctly
	
	self:updateUnitLevel();
	
end

-------------------------------------------------------------------------------

function prototype:updateRaidTarget()

	local unit_id = self.layout.unit_id;
	
	if UnitExists( unit_id ) then
		
		local idx = GetRaidTargetIndex( unit_id );

		if idx ~= self.frame.raid_icon.idx then
		
			self.frame.raid_icon.idx = idx;
			
			if not idx then

				self.frame.raid_icon:SetAlpha( 0 );

			else
				
				local col = ((idx-1) % 4) * 0.25;
				local row = math.floor( (idx-1) / 4 ) * 0.25;
				
				self.frame.raid_icon.texture:SetTexCoord( col, row, col, row+0.25, col+0.25, row, col+0.25, row+0.25 );
				self.frame.raid_icon:SetAlpha( 1 );
				
			end
		end
		
	elseif self.frame.raid_icon.idx then
		
		self.frame.raid_icon.idx = nil;
		self.frame.raid_icon:SetAlpha( 0 );
		
	end
end

-------------------------------------------------------------------------------

function prototype:updateStatus( unit_id )

	if UnitExists( unit_id ) and UnitIsUnit( unit_id, self.layout.unit_id ) then
			
		-- update the unit's state -- AFK, OFFLINE, Feigning Death... etc
		
		local status = nil;
	
		if not UnitIsConnected( unit_id ) then
			if not UnitIsUnit( "pet", unit_id ) then
				status = "OFFLINE";
			end
		elseif UnitIsAFK( unit_id ) then
			status = "AFK";
		elseif UnitIsFeignDeath( unit_id ) then
			status = "FD";
		elseif UnitIsDead( unit_id ) then
			status = "DEAD";
		elseif UnitIsDeadOrGhost( unit_id ) then
			status = "GHOST";
		end
		
		if not status and self.frame.state.status then
			self.frame.state.status = nil;
			self.frame.state:SetAlpha( 0 );
		elseif status ~= self.frame.state.status then
			self.frame.state.status = status;
			self.frame.state.text:SetText( status );
			self.frame.state:SetAlpha( 1 );
		end
	end			
end

-------------------------------------------------------------------------------

function prototype:enterCombat()

	if self.has_model and not self.in_combat then

		local unit_id = self.layout.unit_id;
		
		if UnitIsUnit( unit_id, "player" ) or UnitIsUnit( unit_id, "pet" ) then
				
			local status   = self.frame.model.status;
			
			status.texture:SetTexCoord( 0.5, 0, 0.5, 1, 1, 0, 1, 1 );
			status:SetAlpha( 1 );
			
			
			self.in_combat = true;
		end	
	end	
end

-------------------------------------------------------------------------------

function prototype:leaveCombat()
	
	self.in_combat = false;

	if self.has_model then
			
		local status   = self.frame.model.status;
		
		if status:GetAlpha() ~= 0 then
			self.frame.model.status:SetAlpha( 0 );
		end
	
		self:updateResting();
		
	end		
end

-------------------------------------------------------------------------------

function prototype:updateResting()
	
	if not self.in_combat and self.has_model then
	
		local status = self.frame.model.status;

		if UnitIsUnit( "player", self.layout.unit_id ) and IsResting() then
			
			status.texture:SetTexCoord( 0, 0, 0, 1, 0.5, 0, 0.5, 1 );
			status:SetAlpha( 1 );
			
			
		elseif status:GetAlpha() ~= 0 then
			
			status:SetAlpha( 0 );		
			
		end		
	end	
end

-------------------------------------------------------------------------------
--        PERIODIC STATUS MANAGEMENT -- Called from events typically         --
-------------------------------------------------------------------------------

function prototype:refreshUnit()

	local frame   = self.frame;
	local unit_id = self.layout.unit_id;
	local exists  = UnitExists( unit_id );
	local visible = self.frame:IsVisible();
			
	-- manage changes in who the unit is pointing to if the unit is valid and active
	
	if exists and visible then
		
		-- this handles swapping animation in and out based on whether the unit in
		-- question is currently in or out of visibility range. Generally, WoW only
		-- keeps objects within 100 yards cached... if the unit you're looking at
		-- wanders outside of that range (roughly) their model animation stops
		-- working and you get a blank unit portrait. This test will swap the animation
		-- for a still portrait when they are out of range and back again when they
		-- return to visibility
		
		if frame.model then
			
			if self.layout.Portrait.visible == "anim" then
				
				if frame.model.anim ~= nil and UnitIsVisible( unit_id ) == nil then
					
					self:updateModel();
					
				elseif frame.model.anim == nil and UnitIsVisible( unit_id ) ~= nil then
					
					self:updateModel();
					
				end
			end
		end				
	end
end

-------------------------------------------------------------------------------

function prototype:updateTalentBuild( build )

	local unit_type   = nil;
	local unit_id     = self.layout.unit_id;
	self.unit_spec    = nil;

	if not build and UnitExists( unit_id ) and UnitIsPlayer( unit_id ) then
		nUI_GetTalentBuild( self, self.layout.unit_id );
	end
	
	if build then
			
		local name1, num1 = build.name1 or "", build.num1 or 0;
		local name2, num2 = build.name2 or "", build.num2 or 0;
		local name3, num3 = build.name3 or "", build.num3 or 0;
		
		if num1 > 0 or num2 > 0 or num3 > 0 then
			
			if self.layout.browse_spec == "text" then
	
				if num1 > num2 and num1 > num3 then
					
					unit_type = name1;
					
				elseif num2 > num1 and num2 > num3 then
					
					unit_type = name2;
					
				elseif num3 > num1 and num3 > num2 then
					
					unit_type = name3;
					
				end
			end
			
			if not unit_type then
				
				unit_type = ( "%d / %d /%d" ):format( num1, num2, num3 );
				self.unit_spec = unit_type;
	
			else
				
				self.unit_spec = unit_type.." ( "..( "%d / %d /%d" ):format( num1, num2, num3 ).." )";
				
			end					
		end		
	end
	
--	nUI:debugMsg( "talent build for "..self.layout.unit_id.." = ["..(unit_type or "nil").."]" );
	
	self:setUnitType( unit_type, nil );
	
end

-------------------------------------------------------------------------------

function prototype:updateUnitInfo()

	-- display an icon if the unit is rare, elite or world boss if we're using portraits
	-- if the unit is a player, do spec browsing
	-- display info about the unit

	local unit_id      = self.layout.unit_id;
	
	if UnitExists( unit_id ) then
					
		local texture_file = nil;
		local unit_type    = nil;
		
		self.unit_spec = nil;
	
		if UnitIsPlayer( unit_id ) then
			
			if self.layout.browse_spec ~= "off" 
			and UnitIsVisible( unit_id ) 
			and UnitLevel( unit_id ) > 10 then
				
				self:setUnitType( nil, nil );
				self.unit_spec = "(talent build pending update)";
				
				nUI_GetTalentBuild( self, unit_id );
				
			end	
			
		elseif self.has_model then

			self.classification = UnitClassification( unit_id );
					
			if self.classification == "rare" then
				unit_type     = "Rare";
				self.is_elite = false;
				texture_file  = "Interface\\AddOns\\nUI\\Units\\Art\\nUI_Icon_Rare";
			elseif self.classification == "rareelite" then
				unit_type     = "Rare Elite";
				self.is_elite = true;
				texture_file  = "Interface\\AddOns\\nUI\\Units\\Art\\nUI_Icon_Rare";
			elseif self.classification == "elite" then
				unit_type     = "Elite";
				self.is_elite = true;
				texture_file  = "Interface\\AddOns\\nUI\\Units\\Art\\nUI_Icon_Elite";
			elseif self.classification == "worldboss" then
				unit_type     = "World Boss";
				self.is_elite = true;
				texture_file  = "Interface\\AddOns\\nUI\\Units\\Art\\nUI_Icon_WorldBoss";
			else

				self.is_elite = false;
				
				if self.classification ~= "normal" then
					nUI:debugMsg( 1, "unhandled unit classification ["..self.classification.."]" );
				end
			end
			
			self:setUnitType( unit_type, texture_file );
		end			
	end
end

-------------------------------------------------------------------------------

function prototype:setUnitType( unit_type, texture_file )
	
	if self.has_model then
		if unit_type ~= self.unit_type then
			
			self.unit_type = unit_type;
			
			if unit_type then
				
				self.frame.model.info.text:SetText( unit_type );
				self.frame.model.info:SetAlpha( 1 );
				
			else
				
				self.frame.model.info.text:SetText( "" );
				self.frame.model.info:SetAlpha( 0 );
			end
		end
		
		if self.layout.elite_icon == "yes" then
			
			if texture_file ~= nil then
				
				texture_file = texture_file.."_"..self.layout.Portrait.style.."_"..self.layout.orientation..".blp";
				
				self.frame.model.elite.texture:SetTexture( texture_file );
				self.frame.model.elite:SetAlpha( 1 );
			else
				self.frame.model.elite:SetAlpha( 0 );
			end
		end
	end
end

-------------------------------------------------------------------------------

function prototype:updateUnitClass()

	if UnitExists( self.layout.unit_id ) then		
			
		-- display the unit's class (warrior, rogue, shaman, etc.)
		
		self.localized_class, self.unit_class = UnitClass( self.layout.unit_id or "none" );

		local color = RAID_CLASS_COLORS[self.unit_class];

		if color and self.layout.class_colors ~= "no" then
			self.frame.label.text:SetTextColor( color.r, color.g, color.b, 1 );
		else
			self.frame.label.text:SetTextColor( 1, 0.8196079, 0, 1 );
		end
	
		if self.frame.unit_class then			
		
			local class_row, class_col;
			
			if self.unit_class == "WARRIOR" then
				class_row = 0;
				class_col = 0;
			elseif self.unit_class == "MAGE" then
				class_row = 0;
				class_col = 1;
			elseif self.unit_class == "ROGUE" then
				class_row = 0;
				class_col = 2;
			elseif self.unit_class == "DRUID" then
				class_row = 0;
				class_col = 3;
			elseif self.unit_class == "HUNTER" then
				class_row = 1;
				class_col = 0;
			elseif self.unit_class == "SHAMAN" then
				class_row = 1;
				class_col = 1;
			elseif self.unit_class == "PRIEST" then
				class_row = 1;
				class_col = 2;
			elseif self.unit_class == "WARLOCK" then
				class_row = 1;
				class_col = 3;
			elseif self.unit_class == "PALADIN" then
				class_row = 2;
				class_col = 0;
			else
				nUI:debugMsg( 1, "unhandled unit class ["..(self.unit_class or "unknown").."] in ["..self.layout.unit_id.."]" );
				class_row = 2;
				class_col = 1;
			end
		
			local x1 = class_col * 64;
			local x2 = x1 + 64;
			local y1 = class_row * 64;
			local y2 = y1 + 64;
			
			x1 = x1 / 256;
			x2 = x2 / 256;
			y1 = y1 / 256;
			y2 = y2 / 256;
			
			self.frame.unit_class.texture:SetTexCoord( x1, y1, x1, y2, x2, y1, x2, y2 );
		end
		
	end

end

-------------------------------------------------------------------------------

function prototype:setRaidID( unit_id, id )
	
	if self.layout.unit_id ~= unit_id then
		
		self.layout.unit_id = unit_id;
		self.layout.raid_id = id;
		self.layout.id      = id;
	
		self:initUnitFrame( self.frame, false );
		
		if self.frame.model      then self:initUnitFrame( self.frame, true ); end
		if self.frame.label      then self:initUnitFrame( self.frame.label, true ); end
		if self.frame.level      then self:initUnitFrame( self.frame.level, true ); end
		if self.frame.unit_class then self:initUnitFrame( self.frame.unit_class, true ); end
		if self.frame.status     then self:initUnitFrame( self.frame.status, true ); end
		
		self:checkGUID();
		self:updateRaidInfo();
		
	end	
end